/*
 * fbvnc - a small linux framebuffer vnc viewer
 *
 * Copyright (C) 2009-2011 Ali Gholami Rudi
 *
 * This program is released under GNU GPL version 2.
 */
#include <arpa/inet.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <poll.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <linux/input.h>
#include "draw.h"
#include "vnc.h"

/* framebuffer depth */
typedef unsigned int fbval_t;

/* optimized version of fb_val() */
#define FB_VAL(r, g, b)	fb_val((r), (g), (b))

#define MIN(a, b)	((a) < (b) ? (a) : (b))
#define MAX(a, b)	((a) > (b) ? (a) : (b))

#define VNC_PORT		"5900"

#define MAXRES			(1 << 21)
#define MIN(a, b)		((a) < (b) ? (a) : (b))

static int cols, rows;
static int mr, mc;		/* mouse position */

static char buf[MAXRES];

static int vnc_connect(char *addr, char *port)
{
	struct addrinfo hints, *addrinfo;
	int fd;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;

	if (getaddrinfo(addr, port, &hints, &addrinfo))
		return -1;
	fd = socket(addrinfo->ai_family, addrinfo->ai_socktype,
			addrinfo->ai_protocol);

	if (connect(fd, addrinfo->ai_addr, addrinfo->ai_addrlen) == -1) {
		close(fd);
		freeaddrinfo(addrinfo);
		return -1;
	}
	freeaddrinfo(addrinfo);
	return fd;
}

static int vnc_init(int fd)
{
	char vncver[] = "RFB 003.003\n";
	struct vnc_client_init clientinit;
	struct vnc_server_init serverinit;
	struct vnc_client_pixelfmt pixfmt_cmd;
	int connstat = VNC_CONN_FAILED;

	write(fd, vncver, 12);
	read(fd, vncver, 12);

	read(fd, &connstat, sizeof(connstat));

	if (ntohl(connstat) != VNC_CONN_NOAUTH)
		return -1;

	clientinit.shared = 1;
	write(fd, &clientinit, sizeof(clientinit));
	read(fd, &serverinit, sizeof(serverinit));

	if (fb_init())
		return -1;
	if (FBM_BPP(fb_mode()) != sizeof(fbval_t)) {
		fprintf(stderr, "fbvnc: fbval_t doesn't match fb depth\n");
		exit(1);
	}
	cols = MIN(ntohs(serverinit.w), fb_cols());
	rows = MIN(ntohs(serverinit.h), fb_rows());
	mr = rows / 2;
	mc = cols / 2;

	read(fd, buf, ntohl(serverinit.len));
	pixfmt_cmd.type = VNC_CLIENT_PIXFMT;
	pixfmt_cmd.format.bpp = 8;
	pixfmt_cmd.format.depth = 8;
	pixfmt_cmd.format.bigendian = 0;
	pixfmt_cmd.format.truecolor = 1;

	pixfmt_cmd.format.rmax = htons(3);
	pixfmt_cmd.format.gmax = htons(7);
	pixfmt_cmd.format.bmax = htons(7);
	pixfmt_cmd.format.rshl = 0;
	pixfmt_cmd.format.gshl = 2;
	pixfmt_cmd.format.bshl = 5;

	write(fd, &pixfmt_cmd, sizeof(pixfmt_cmd));
	return fd;
}

static int vnc_free(void)
{
	fb_free();
	return 0;
}

static int vnc_refresh(int fd, int inc)
{
	struct vnc_client_fbup fbup_req;
	fbup_req.type = VNC_CLIENT_FBUP;
	fbup_req.inc = inc;
	fbup_req.x = htons(0);
	fbup_req.y = htons(0);
	fbup_req.w = htons(cols);
	fbup_req.h = htons(rows);
	write(fd, &fbup_req, sizeof(fbup_req));
	return 0;
}

static void drawfb(char *s, int x, int y, int w, int h)
{
	fbval_t slice[1 << 14];
	int i, j;
	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j++) {
			int c = *(unsigned char *) &s[i * w + j];
			int r = (c & 0x3) << 6;
			int g = ((c >> 2) & 0x7) << 5;
			int b = ((c >> 5) & 0x7) << 5;
			slice[j] = FB_VAL(r, g, b);
		}
		fb_set(y + i, x, slice, w);
	}
}

static void xread(int fd, void *buf, int len)
{
	int nr = 0;
	int n;
	while (nr < len && (n = read(fd, buf + nr, len - nr)) > 0)
		nr += n;
	if (nr < len) {
		printf("partial vnc read!\n");
		exit(1);
	}
}

static int vnc_event(int fd)
{
	struct vnc_rect uprect;
	char msg[1 << 12];
	struct vnc_server_fbup *fbup = (void *) msg;
	struct vnc_server_cuttext *cuttext = (void *) msg;
	struct vnc_server_colormap *colormap = (void *) msg;
	int j;
	int n;

	if (read(fd, msg, 1) != 1)
		return -1;
	switch (msg[0]) {
	case VNC_SERVER_FBUP:
		xread(fd, msg + 1, sizeof(*fbup) - 1);
		n = ntohs(fbup->n);
		for (j = 0; j < n; j++) {
			int x, y, w, h;
			xread(fd, &uprect, sizeof(uprect));
			x = ntohs(uprect.x);
			y = ntohs(uprect.y);
			w = ntohs(uprect.w);
			h = ntohs(uprect.h);
			if (x >= cols || x + w > cols)
				return -1;
			if (y >= rows || y + h > rows)
				return -1;
			xread(fd, buf, w * h);
			drawfb(buf, x, y, w, h);
		}
		break;
	case VNC_SERVER_BELL:
		break;
	case VNC_SERVER_CUTTEXT:
		xread(fd, msg + 1, sizeof(*cuttext) - 1);
		xread(fd, buf, ntohl(cuttext->len));
		break;
	case VNC_SERVER_COLORMAP:
		xread(fd, msg + 1, sizeof(*colormap) - 1);
		xread(fd, buf, ntohs(colormap->n) * 3 * 2);
		break;
	default:
		fprintf(stderr, "unknown vnc msg: %d\n", msg[0]);
		return -1;
	}
	return 0;
}

static int rat_event(int fd, int ratfd)
{
	char ie[3];
	struct vnc_client_ratevent me = {VNC_CLIENT_RATEVENT};
	int mask = 0;
	if (read(ratfd, &ie, sizeof(ie)) != 3)
		return -1;
	mc += ie[1];
	mr -= ie[2];
	mc = MAX(0, MIN(cols - 1, mc));
	mr = MAX(0, MIN(rows - 1, mr));
	if (ie[0] & 0x01)
		mask |= VNC_BUTTON1_MASK;
	if (ie[0] & 0x04)
		mask |= VNC_BUTTON2_MASK;
	if (ie[0] & 0x02)
		mask |= VNC_BUTTON3_MASK;
	me.y = htons(mr);
	me.x = htons(mc);
	me.mask = mask;
	write(fd, &me, sizeof(me));
	return 0;
}

static int press(int fd, int key, int down)
{
	struct vnc_client_keyevent ke = {VNC_CLIENT_KEYEVENT};
	ke.key = htonl(key);
	ke.down = down;
	return write(fd, &ke, sizeof(ke));
}

static int kbd_event(int fd, int kbdfd)
{
	char key[1024];
	int i, nr;

	if ((nr = read(kbdfd, key, sizeof(key))) <= 0 )
		return -1;
	for (i = 0; i < nr; i++) {
		int k = -1;
		int mod[4];
		int nmod = 0;
		switch (key[i]) {
		case 0x08:
		case 0x7f:
			k = 0xff08;
			break;
		case 0x09:
			k = 0xff09;
			break;
		case 0x1b:
			if (i + 2 < nr && key[i + 1] == '[') {
				if (key[i + 2] == 'A')
					k = 0xff52;
				if (key[i + 2] == 'B')
					k = 0xff54;
				if (key[i + 2] == 'C')
					k = 0xff53;
				if (key[i + 2] == 'D')
					k = 0xff51;
				if (key[i + 2] == 'H')
					k = 0xff50;
				if (k > 0) {
					i += 2;
					break;
				}
			}
			k = 0xff1b;
			if (i + 1 < nr) {
				mod[nmod++] = 0xffe9;
				k = key[++i];
				if (k == 0x03)	/* esc-^C: quit */
					return -1;
			}
			break;
		case 0x0d:
			k = 0xff0d;
			break;
		case 0x0c:	/* ^L: redraw */
			if (vnc_refresh(fd, 0))
				return -1;
		default:
			k = (unsigned char) key[i];
		}
		if (k >= 'A' && k <= 'Z' || strchr(":\"<>?{}|+_()*&^%$#@!~", k))
			mod[nmod++] = 0xffe1;
		if (k >= 1 && k <= 26) {
			k = 'a' + k - 1;
			mod[nmod++] = 0xffe3;
		}
		if (k > 0) {
			int j;
			for (j = 0; j < nmod; j++)
				press(fd, mod[j], 1);
			press(fd, k, 1);
			press(fd, k, 0);
			for (j = 0; j < nmod; j++)
				press(fd, mod[j], 0);
		}
	}
	return 0;
}

static void term_setup(struct termios *ti)
{
	struct termios termios;
	char *hide = "\x1b[?25l";
	char *clear = "\x1b[2J\x1b[H";
	char *msg = "\t\t\t*** fbvnc ***\r\n";

	write(STDIN_FILENO, hide, strlen(hide));
	write(STDOUT_FILENO, clear, strlen(clear));
	write(STDOUT_FILENO, msg, strlen(msg));
	tcgetattr (0, &termios);
	*ti = termios;
	cfmakeraw(&termios);
	tcsetattr(0, TCSANOW, &termios);
}

static void term_cleanup(struct termios *ti)
{
	char *show = "\x1b[?25h";
	tcsetattr(0, TCSANOW, ti);
	write(STDIN_FILENO, show, strlen(show));
}

static void mainloop(int vnc_fd, int kbd_fd, int rat_fd)
{
	struct pollfd ufds[3];
	int pending = 0;
	int err;
	ufds[0].fd = kbd_fd;
	ufds[0].events = POLLIN;
	ufds[1].fd = vnc_fd;
	ufds[1].events = POLLIN;
	ufds[2].fd = rat_fd;
	ufds[2].events = POLLIN;
	if (vnc_refresh(vnc_fd, 0))
		return;
	while (1) {
		err = poll(ufds, 3, 500);
		if (err == -1 && errno != EINTR)
			break;
		if (!err)
			continue;
		if (ufds[0].revents & POLLIN)
			if (kbd_event(vnc_fd, kbd_fd) == -1)
				break;
		if (ufds[1].revents & POLLIN) {
			if (vnc_event(vnc_fd) == -1)
				break;
			pending = 0;
		}
		if (ufds[2].revents & POLLIN)
			if (rat_event(vnc_fd, rat_fd) == -1)
				break;
		if (!pending++)
			if (vnc_refresh(vnc_fd, 1))
				break;
	}
}

int main(int argc, char * argv[])
{
	char *port = VNC_PORT;
	char *host = "127.0.0.1";
	struct termios ti;
	int vnc_fd, rat_fd;
	if (argc >= 2)
		host = argv[1];
	if (argc >= 3)
		port = argv[2];
	if ((vnc_fd = vnc_connect(host, port)) == -1) {
		fprintf(stderr, "could not connect!\n");
		return 1;
	}
	if (vnc_init(vnc_fd) == -1) {
		fprintf(stderr, "vnc init failed!\n");
		return 1;
	}
	term_setup(&ti);
	rat_fd = open("/dev/input/mice", O_RDONLY);

	mainloop(vnc_fd, 0, rat_fd);

	term_cleanup(&ti);
	vnc_free();
	close(vnc_fd);
	close(rat_fd);
	return 0;
}
